import { HttpModule } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { getModelToken } from '@nestjs/mongoose';
import { Test } from '@nestjs/testing';
import * as ejs from 'ejs';
import { MailerModule } from '../mailer/mailer.module';
import { MailerService } from '../mailer/mailer.service';
import { TempUserService } from './temp-user.service';
import { Types } from 'mongoose';

const ejsSpy = jest.spyOn(ejs, 'renderFile');

describe('TempUserService', () => {
  let tempUserService: TempUserService;

  const tempUserModelMock = {
    create: jest.fn(),
    findOne: jest.fn(),
    findById: jest.fn(),
    findByIdAndUpdate: jest.fn(),
    deleteOne: jest.fn(),
    find: jest.fn().mockReturnThis(),
    exec: jest.fn(),
    updateOne: jest.fn(),
    populate: jest.fn().mockReturnThis(),
  };

  const mockMailService = {
    send: jest.fn(),
    getTemplateLocation: jest.fn(),
    loadJsonConfig: jest.fn().mockReturnValue({ subject: 'subject' }),
    config: {
      templates: {
        userAddedToStructure: {
          ejs: 'userAddedToStructure.ejs',
          json: 'userAddedToStructure.json',
        },
        tempUserRegistration: {
          ejs: 'tempUserRegistration.ejs',
          json: 'tempUserRegistration.json',
        },
      },
    },
  };

  // Set up the test module before each test
  beforeEach(async () => {
    const module = await Test.createTestingModule({
      imports: [HttpModule, MailerModule],
      providers: [
        TempUserService,
        {
          provide: getModelToken('TempUser'),
          useValue: tempUserModelMock,
        },
        { provide: MailerService, useValue: mockMailService },
      ],
    }).compile();

    tempUserService = module.get<TempUserService>(TempUserService);
  });

  // Test whether the service is defined
  it('should be defined', () => {
    expect(tempUserService).toBeDefined();
  });

  // Test related to the 'findAll' method
  describe('findAll', () => {
    it('should return a list of temp users', async () => {
      const tempUsers = [
        { email: 'user1@test.com', structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')] },
        { email: 'user2@test.com', structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')] },
      ];

      tempUserModelMock.exec.mockResolvedValueOnce(tempUsers);

      expect(await tempUserService.findAll()).toEqual(tempUsers);
    });
  });

  // Tests related to the 'create' method
  describe('create', () => {
    const tmpUser = {
      email: 'test@test.com',
      structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
    };

    it('should not create temporary user if it already exists', async () => {
      tempUserModelMock.findOne.mockReturnThis();
      tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
      try {
        await tempUserService.create(tmpUser, 'PIMMS Framboise');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('User already exists');
        expect(e.status).toEqual(HttpStatus.BAD_REQUEST);
      }
    });

    it('should create a temporary user', async () => {
      tempUserModelMock.findOne.mockReturnThis();
      tempUserModelMock.exec.mockResolvedValueOnce(null);
      tempUserModelMock.create.mockResolvedValueOnce(tmpUser);
      tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
      ejsSpy.mockResolvedValue('mock');
      expect(await tempUserService.create(tmpUser, 'PIMMS Framboise')).toEqual(tmpUser);
      expect(mockMailService.send).toHaveBeenCalled();
    });
  });

  // Test whether the 'findOne' method works
  it('should find one', async () => {
    const tmpUser = {
      email: 'test2@test.com',
      structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
    };
    tempUserModelMock.findOne.mockReturnThis();
    tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
    expect(await tempUserService.findOne('test2@test.com')).toEqual(tmpUser);
  });

  // Test whether the 'findById' method works
  it('should find one by id', async () => {
    const tmpUser = {
      email: 'test2@test.com',
      structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
    };
    tempUserModelMock.findById.mockReturnThis();
    tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
    expect(await tempUserService.findById('5fbb92e480a5c257dc0161f0')).toEqual(tmpUser);
  });

  // Tests related to the 'delete' method
  describe('delete', () => {
    it('should delete a temporary user', async () => {
      const tmpUser = {
        email: 'test2@test.com',
        structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
      };
      tempUserModelMock.findOne.mockReturnThis();
      tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
      tempUserModelMock.deleteOne.mockReturnThis();
      tempUserModelMock.exec.mockImplementationOnce(() => ({}));
      expect(await tempUserService.delete('toto@test.com')).toEqual(tmpUser);
    });

    it('should return an error if the user does not exist', async () => {
      tempUserModelMock.findOne.mockReturnThis();
      tempUserModelMock.exec.mockResolvedValueOnce(null);
      try {
        await tempUserService.delete('toto@test.com');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('User does not exist');
        expect(e.status).toEqual(HttpStatus.BAD_REQUEST);
      }
    });
  });

  // Tests related to the 'updateStructureLinked' method
  describe('updateStructureLinked', () => {
    it('should update the linked structure', async () => {
      const tmpUser = {
        email: 'test2@test.com',
        structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
      };
      tempUserModelMock.findOne.mockReturnThis();
      tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);

      const userInDb = {
        structuresLink: [],
        save: jest.fn().mockResolvedValueOnce(tmpUser),
      };

      jest.spyOn(tempUserModelMock, 'findOne').mockReturnValue({
        exec: jest.fn().mockResolvedValueOnce(userInDb),
      });

      tempUserModelMock.findByIdAndUpdate.mockReturnThis();
      expect(await tempUserService.updateStructureLinked(tmpUser)).toEqual(tmpUser);
    });

    it('should not update the linked structure if the user is already linked', async () => {
      const tmpUser = {
        email: 'test2@test.com',
        structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
      };
      tempUserModelMock.findOne.mockReturnThis();

      const userInDb = {
        structuresLink: [new Types.ObjectId()],
        save: jest.fn().mockResolvedValueOnce(tmpUser),
      };

      jest.spyOn(tempUserModelMock, 'findOne').mockReturnValue({
        exec: jest.fn().mockResolvedValueOnce(userInDb),
      });

      tempUserModelMock.findByIdAndUpdate.mockReturnThis();

      try {
        await tempUserService.updateStructureLinked(tmpUser);
      } catch (e) {
        expect(e.message).toEqual('User is already linked');
        expect(e.status).toEqual(HttpStatus.UNPROCESSABLE_ENTITY);
      }
    });
  });
});
